/**
 * Abbaye des Morts for Parallax Propeller
 * Copyright (c) 2015 Marco Maccaferri
 *
 * TERMS OF USE: MIT License
 */

#include <propeller.h>

#include "i2c_driver.h"

#define EEPROM_ADDR 0xA0

static uint32_t scl = 1 << I2C_SCL;
static uint32_t sda = 1 << I2C_SDA;
static uint32_t sda_shift = I2C_SDA;

void i2c_initialize() {
    int32_t i;

    OUTA |= scl;                        // An I2C device may be left in an invalid state and may need to be
    DIRA |= scl;                        // reinitialized.  Drive SCL high.
    DIRA &= ~sda;                       // Set SDA as input
    for (i = 0; i < 9; i++) {
        OUTA &= ~scl;                   // Put out up to 9 clock pulses
        OUTA |= scl;
        if ((INA & sda) != 0)           // Repeat if SDA not driven high by the EEPROM
            break;
    }
}

// SDA goes HIGH to LOW with SCL HIGH
void i2c_start() {
    OUTA |= scl;                        // Initially drive SCL HIGH
    DIRA |= scl;
    OUTA |= sda;                        // Initially drive SDA HIGH
    DIRA |= sda;
    OUTA &= ~sda;                       // Now drive SDA LOW
    OUTA &= ~scl;                       // Leave SCL LOW
}

// SDA goes LOW to HIGH with SCL High
void i2c_stop() {
    OUTA |= scl;                        // Drive SCL HIGH
    OUTA |= sda;                        // then SDA HIGH
    DIRA &= ~scl;                       // Now let them float
    DIRA &= ~sda;                       // If pullups present, they'll stay HIGH
}

// Write i2c data.  Data byte is output MSB first, SDA data line is valid
// only while the SCL line is HIGH.  Data is always 8 bits (+ ACK/NAK).
// SDA is assumed LOW and SCL and SDA are both left in the LOW state.
int32_t i2c_write(uint8_t data) {
    int32_t i, ack = 0;

    for (i = 0x80; i != 0x00; i >>= 1) {    // Output data to SDA
        if ((data & i) != 0)
            OUTA |= sda;
        else
            OUTA &= ~sda;
        OUTA |= scl;                        // Toggle SCL from LOW to HIGH to LOW
        OUTA &= ~scl;
    }

    DIRA &= ~sda;                           // Set SDA to input for ACK/NAK
    OUTA |= scl;
    ack = INA & sda;                        // Sample SDA when SCL is HIGH
    OUTA &= ~scl;

    OUTA &= ~sda;                           // Leave SDA driven LOW
    DIRA |= sda;

    return ack == 0 ? I2C_ACK : I2C_NAK;
}

// Read in i2c data, Data byte is output MSB first, SDA data line is
// valid only while the SCL line is HIGH.  SCL and SDA left in LOW state.
uint8_t i2c_read(int32_t ack) {
    int32_t i, data = 0;

    DIRA &= ~sda;                       // Make SDA an input
    for (i = 0; i < 8; i++) {           // Receive data from SDA
        OUTA |= scl;                    // Sample SDA when SCL is HIGH
        data <<= 1;
        data |= (INA & sda) >> sda_shift;
        OUTA &= ~scl;
    }

    if (ack)                            // Output ACK/NAK to SDA
        OUTA |= sda;
    else
        OUTA &= ~sda;
    DIRA |= sda;
    OUTA |= scl;                        // Toggle SCL from LOW to HIGH to LOW
    OUTA &= ~scl;

    OUTA &= ~sda;                       // Leave SDA driven LOW

    return data;
}

int32_t i2c_read_page(uint32_t addr, uint8_t * ptr, uint32_t count) {

    while (count > 0) {
        i2c_start();                                    // Select the device & send address
        i2c_write(EEPROM_ADDR | I2C_WRITE | ((addr & 0x10000) >> 13));
        i2c_write(addr >> 8);
        i2c_write(addr & 0xFF);

        i2c_start();                                    // Reselect the device for reading
        i2c_write(EEPROM_ADDR | I2C_READ | ((addr & 0x10000) >> 13));

        addr++;
        while (count > 1 && (addr & 0xFFFF) != 0) {
            *ptr++ = i2c_read(I2C_ACK);
            count--;
            addr++;
        }
        *ptr++ = i2c_read(I2C_NAK);
        count--;

        i2c_stop();
    }

    return 0;
}
